# MIT License
#
# Copyright (c) 2020, Niklas Rosenstein
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# this software and associated documentation files (the "Software"), to deal in
# Software without restriction, including without limitation the rights to use,
# modify, merge, publish, distribute, sublicense, and/or sell copies of the
# and to permit persons to whom the Software is furnished to do so, subject to
# following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
# PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR
# USE OR OTHER DEALINGS IN THE SOFTWARE.

import datetime
import enum
import flask
import logging
import nr.config
import os
import shlex
import subprocess as sp
import yaml

from cube import spec
from cube.client import api
from cube.daemon.api import spec_context, config, mapper
from cube.daemon.config import DatabaseConfig
from cube.daemon.utils.logging import class_logger
from nr.interface import implements
from nr.parsing.date import timezone
from nr.utils.ponyorm.converters import EnumConverter
from nr.utils.ponyorm.entity import override_entity_wrapper
from pony import orm
from typing import Iterable, List, Optional

db = orm.Database()
commit = orm.commit
session = orm.db_session
select = orm.select


def initialize(config: DatabaseConfig):
  if config.filename and config.options.get('create_db'):
    os.makedirs(os.path.dirname(config.filename), exist_ok=True)
  db.bind(**config.bind_kwargs())
  EnumConverter.register(db)
  db.generate_mapping(create_tables=True)
  logging.getLogger(__name__).info('Initialized database.')


def datetime_field(**kwargs):
  if 'default' not in kwargs:
    kwargs['default'] = lambda: datetime.datetime.now(timezone.utc)
  return orm.Required(datetime.datetime, **kwargs)


def repair_datetime(dt: datetime.datetime) -> datetime.datetime:
  return dt.replace(tzinfo=timezone.utc) if dt and not dt.tzinfo else dt


def untrack(value):
  """
  PonyORM's #orm.Json field type returns "tracked" dicts/lists, which cannot
  be handled by the YAML serializer. This function converts those tracked
  versions to normal dicts/lists.
  """

  if isinstance(value, orm.ormtypes.TrackedDict):
    value = {k: untrack(v) for k, v in value.items()}
  elif isinstance(value, orm.ormtypes.TrackedList):
    value = [untrack(v) for v in value]
  return value


@implements(spec.IRoleDiscovery)
class RoleDiscoverer:
  """
  Discover roles stored in the database. Requires that it is used inside
  a #session.
  """

  def __init__(self, exclude_services: Iterable[str]):
    self.exclude_services = frozenset(exclude_services)

  def discover_roles(self, roles: List[str]) -> Iterable[api.AbstractConsumedRole]:
    query = db.ProducedRole.select(lambda r: r.role in roles and
      r.service.name not in self.exclude_services)
    for role in query:
      yield role.to_consumed_role()


class Product(db.Entity):
  #: The name of the product. This, together with the #version
  #: uniquely identifies the product.
  name = orm.Required(str)

  #: The product version number.
  version = orm.Required(str)

  #: The status of the product. This is set from the product installation
  #: task that is run in the background when a new product is installed.
  status = orm.Required(api.ProductStatus, default=api.ProductStatus.CREATED)

  #: A string that describes where the Product was installed from.
  #: This is not necessarily machine-processable.
  origin = orm.Required(str)

  #: A JSON payload that deserializes into the base configuration of
  #: the product. This is not always defined, for example if the
  #: product is installed on the host directory for development purposes.
  #: This field exists purely for performance reasons if we need to render
  #: a service's full configuration.
  config = orm.Optional(orm.Json)

  #: True if the #config should be cached or not.
  cache_config = orm.Required(bool, default=True)

  #: The path where the product is installed on the host.
  path = orm.Optional(str, unique=True)

  #: The date at which the product was installed.
  installed_at = datetime_field()

  orm.PrimaryKey(name, version)
  services = orm.Set('Service')

  def to_api(self) -> api.ProductDetails:
    return api.ProductDetails(
      name=self.name, version=self.version, status=self.status,
      origin=self.origin, installed_at=repair_datetime(self.installed_at),
      installed_for=list(orm.select(s.name for s in self.services)))

  def to_spec(self) -> spec.Product:
    return spec_context.product(self.name, self.version, self.path)

  def get_config(self) -> api.ProductConfig:
    if self.cache_config:
      return mapper.deserialize(self.config, api.ProductConfig)
    else:
      return self.to_spec().get_config()


class Service(db.Entity):
  logger = class_logger()

  #: The unique name of the service.
  name = orm.PrimaryKey(str)

  #: The status of the service.
  status = orm.Required(api.ServiceStatus, default=api.ServiceStatus.CREATED)

  #: The product that is currently (expected) to be installed for this
  #: service. This is optional as a service is not tied to a product
  #: right after installation.
  installed_product = orm.Optional(Product)

  #: The time at which the service was created.
  created_at = datetime_field()

  #: The last time at which the status for the service was updated.
  status_updated_at = orm.Optional(datetime.datetime)

  #: The home directory of the service.
  home = orm.Required(str, unique=True)

  #: The list of role names that are consumed by the product installed for
  #: this service. This is stored as a cache for more effective querying.
  consumed_roles = orm.Required(orm.StrArray, default=lambda: [])

  config_revisions = orm.Set('ServiceConfigRevision')
  produces = orm.Set('ProducedRole')

  def to_api(self) -> api.ServiceDetails:
    return api.ServiceDetails(
      name=self.name,
      installed_product=self.installed_product.to_api() if
        self.installed_product else None,
      created_at=repair_datetime(self.created_at),
      status=self.status,
      status_updated_at=repair_datetime(self.status_updated_at))

  def to_spec(self) -> spec.Service:
    product = self.installed_product.to_spec() if self.installed_product else None
    return spec_context.service(self.name, self.home, product)

  def update_status(self, commit: bool = False) -> api.ServiceStatus:
    if not self.home:
      status = api.ServiceStatus.CREATED
    else:
      try:
        status = self.to_spec().get_status()
      except Exception as exc:
        self.logger.exception('Exception when trying to retrieve the status of '
          'service %r (installed_product: %r)', self, self.installed_product)
        status = api.ServiceStatus.UNKNOWN
    self.status = status
    self.status_updated_at = datetime.datetime.now(timezone.utc)
    if commit:
      orm.commit()
    return status

  def latest_config_revision(self) -> 'Optional[ServiceConfigRevision]':
    return self.config_revisions.select().order_by(
      lambda cr: orm.desc(cr.revision)).first()

  def latest_config_overrides(self) -> api.ConfigOverride:
    revision = self.latest_config_revision()
    if not revision:
      return api.ConfigOverride()
    return revision.to_api().value

  def get_runtime_config(self) -> api.RuntimeConfig:
    product_config = self.installed_product.get_config() if \
      self.installed_product else api.ProductConfig()
    config_overrides = self.latest_config_overrides()
    return spec.generate_runtime_config(
      self.name,
      product_config,
      config_overrides,
      RoleDiscoverer(exclude_services=[self.name]))

  def publish_produced_roles(self):
    """
    Regenerates all #ProducedRole#s associated with this service.
    """

    self.produces.select().delete()
    for role in self.get_runtime_config().produces:
      ProducedRole(service=self, role=role.role, data=role.data)

  def __update_consumed_roles(self):
    old = self.consumed_roles
    if self.installed_product:
      self.consumed_roles = self.installed_product.get_config().deployment.consumed_roles
    else:
      self.consumed_roles = []
    if old != self.consumed_roles:
      self.logger.info('Updated consumed_roles for service %r: %r -> %r.',
        self.name, old, self.consumed_roles)

  def before_insert(self):
    self.__update_consumed_roles()
    # Make sure there exists at least one ServiceConfigRevision.
    if not self.latest_config_revision():
      ServiceConfigRevision(service=self, revision=1, value={})

  def before_update(self):
    self.__update_consumed_roles()


class ServiceConfigRevision(db.Entity):
  #: The service for which this is a configuration.
  service = orm.Required(Service)

  #: The revision number (strictly increasing per service).
  revision = orm.Required(int)

  #: The time at which this config revision was created.
  created_at = datetime_field()

  #: A JSON payload for the config override which must deserialize into
  #: #api.ConfigOverride.
  value = orm.Required(orm.Json)

  orm.PrimaryKey(service, revision)

  def to_api(self) -> api.ConfigOverrideRevision:
    return api.ConfigOverrideRevision(
      service=self.service.name,
      revision=self.revision,
      created_at=repair_datetime(self.created_at),
      value=mapper.deserialize(self.value, api.ConfigOverride))


class ProducedRoleSet(orm.core.SetInstance):

  def consumed_by(self) -> List[Service]:
    """
    Returns the list of services that consume any of the roles in this set.
    """

    # TODO (@NiklasRosenstein): Is there a pure SQL way of querying this?
    services = []
    for role_name in orm.select(p.role for p in self):
      for service in Service.select(lambda s: role_name in s.consumed_roles):
        if service not in services:
          services.append(service)
    return services


class ProducedRole(db.Entity):
  """
  Represents a role that is produced by a product configuration or
  service configuration override. It is stored so that other services
  can efficiently discover roles produced by other services.

  The produced roles for each service are re-created from scratch every
  time the configuration of a service is updated.
  """

  override_entity_wrapper(orm.Set, ProducedRoleSet)

  #: The service that produces this role.
  service = orm.Required(Service)

  #: The role name.
  role = orm.Required(str)

  #: The role data.
  data = orm.Required(orm.Json)

  def to_consumed_role(self) -> api.AbstractConsumedRole:
    return api.AbstractConsumedRole(self.role, untrack(self.data), self.service.name)
