# MIT License
#
# Copyright (c) 2020, Niklas Rosenstein
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# this software and associated documentation files (the "Software"), to deal in
# Software without restriction, including without limitation the rights to use,
# modify, merge, publish, distribute, sublicense, and/or sell copies of the
# and to permit persons to whom the Software is furnished to do so, subject to
# following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
# PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR
# USE OR OTHER DEALINGS IN THE SOFTWARE.

from cube.daemon.config import AuthConfig
from flask import Flask, abort, request
from netaddr import IPAddress, IPNetwork


class AuthMiddleware:
  """
  This middleware registers a `before_request` handler that authenticates a user
  either by validating that the address that they access from is in the
  #AuthConfig.ip_whitelist or that they provide the #AuthConfig.secret in the
  `Authorization` header in the form of `Bearer <secret>`.
  """

  def __init__(self, app: Flask, config: AuthConfig):
    app.before_request(self.before_request)
    self.config = config

  def check_access(self, remote_addr: IPAddress):
    if any(remote_addr in network for network in self.config.ip_whitelist):
      return True
    if self.config.secret:
      auth_header = request.headers.get('Authorization')
      if auth_header == ('Bearer ' + self.config.secret):
        return True
      return False
    return True

  def before_request(self):
    if self.config.real_ip_header and self.config.real_ip_header in request.headers:
      # We assume that this header cannot be forged by a valid client if the
      # service is behind a reverse proxy.
      remote_addr = IPAddress(request.headers[self.config.real_ip_header])
    else:
      remote_addr = IPAddress(request.remote_addr)
    if not self.check_access(remote_addr):
      abort(403)
