# MIT License
#
# Copyright (c) 2020, Niklas Rosenstein
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# this software and associated documentation files (the "Software"), to deal in
# Software without restriction, including without limitation the rights to use,
# modify, merge, publish, distribute, sublicense, and/or sell copies of the
# and to permit persons to whom the Software is furnished to do so, subject to
# following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
# PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR
# USE OR OTHER DEALINGS IN THE SOFTWARE.

from .services import get_service
from cube.client import api
from cube.daemon.api import db, mapper
from cube.daemon.utils.logging import class_logger
from nr.interface import implements


@implements(api.ConfigService)
class ConfigServiceImpl:

  logger = class_logger()

  @db.session
  def get_runtime_config(self, service, revision):
    service = get_service(service)
    return service.get_runtime_config()

  def get_runtime_config_on_disk(self, service):
    raise NotImplementedError

  @db.session
  def get_config_overrides(self, service, revision):
    service = get_service(service)
    if revision is None:
      data = service.latest_config_revision()
    else:
      data = db.ServiceConfigRevision.get(service=service, revision=revision)
    if not data:
      raise api.ResourceDoesNotExist(type='configRevision',
        serviceName=service.name, configRevision=revision)
    return data.to_api()

  def get_config_history(self, service, page_start):
    raise NotImplementedError

  def set_config(self, service, config):
    with db.session:
      service = get_service(service)
      latest_revision = service.latest_config_revision()
      revision_number = latest_revision.revision + 1 if latest_revision else 1
      revision = db.ServiceConfigRevision(
        service=service,
        revision=revision_number,
        value=mapper.serialize(config, api.ConfigOverride))

      self.logger.info('Updated config overrides of service %r (revision: %d).',
        service.name, revision.revision)

    # TODO (@NiklasRosenstein): Unsure why but the above and below cannot be
    #   in the same transaction as otherwise it results in an
    #   UnrepeatableReadError because it things the date of the revision was
    #   changed by another transaction.
    with db.session:
      service = get_service(service.name)
      self.logger.info('Publishing produced roles of service %r.', service.name)
      service.publish_produced_roles()

      # Find all services that consume roles produced by this service in order
      # to update their runtime configuration.
      update_services = service.produces.consumed_by()
      if service not in update_services:
        update_services.append(service)

      config_updates = [
        (s.to_spec(), s.get_runtime_config())
        for s in update_services]

    # TODO (@NiklasRosenstein): Wrap the update in a plan/task.
    for spec_service, runtime_config in config_updates:
      self.logger.info('Updating runtime configuration of service %r', spec_service.name)
      spec_service.update_config(runtime_config)

    return revision.to_api()

  @db.session
  def force_refresh_config(self, service):
    service = get_service(service)
    if not service.installed_product:
      api.IllegalState('Service has no product installed.', serviceName=service.name)
    runtime_config = service.get_runtime_config()
    service.to_spec().update_config(runtime_config)
    return runtime_config
