# MIT License
#
# Copyright (c) 2020, Niklas Rosenstein
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# this software and associated documentation files (the "Software"), to deal in
# Software without restriction, including without limitation the rights to use,
# modify, merge, publish, distribute, sublicense, and/or sell copies of the
# and to permit persons to whom the Software is furnished to do so, subject to
# following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
# PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR
# USE OR OTHER DEALINGS IN THE SOFTWARE.

import copy
import logging
import os
import subprocess
import yaml

from netaddr import IPNetwork
from nr.databind.core import Field, FieldName, Struct, ObjectMapper, PythonClassType
from nr.databind.core.decorations import Remainder, Validator
from nr.databind.json import JsonModule, JsonSerializer
from typing import Optional

logger = logging.getLogger(__name__)


def _canonical(path: Optional[str], parent: str) -> Optional[str]:
  if path is not None:
    path = os.path.abspath(os.path.normpath(os.path.join(parent, path)))
  return path


IPNetworkType = PythonClassType(IPNetwork,
  JsonSerializer(deserialize=lambda mapper, node: IPNetwork(node.value)))


class AuthConfig(Struct):
  #: The name of the header of the users real IP address. We assume that this
  #: header cannot be forged by the client if Cube-Daemon is behind a reverse
  #: proxy.
  real_ip_header = Field(str, default='X-Real-IP')

  #: A list of IPs that have whitelisted access to Cube-Daemon without
  #: requiring further authentication.
  ip_whitelist = Field([IPNetworkType], FieldName('ip-whitelist'))

  #: A secret that is sent in the request headers of the Rest API to gain
  #: access to the resources. If no secret is specified, access is granted
  #: to every client.
  secret = Field(str)


class SslConfig(Struct):
  auto = Field(bool, default=False)
  cert = Field(str, default=False)
  key = Field(str, default=False)

  def get_ssl_context(self, cache_dir: str):
    if self.cert or self.key:
      if not self.cert or not self.key:
        raise RuntimeError('cert and key must be provided')
      return (self.cert, self.key)
    if self.auto:
      cert, key = os.path.join(cache_dir, 'cert.pem'), os.path.join(cache_dir, 'key.pem')
      if not os.path.isfile(cert) or not os.path.isfile(key):
        os.makedirs(cache_dir, exist_ok=True)
        command = ['openssl', 'req', '-new', '-newkey', 'rsa:4096',
          '-days', '365', '-nodes', '-x509', '-subj', '/C=/ST=/L=/O=/CN=localhost',
          '-keyout', key, '-out', cert]
        logger.info('Generating new self-signed certificate with command %r.', command)
        subprocess.check_call(command)
      return (cert, key)
    return None


class ServerConfig(Struct):
  host = Field(str, default='localhost')
  port = Field(int, default=6123)
  ssl = Field(SslConfig, default=SslConfig(auto=True))


class DatabaseConfig(Struct):
  provider = Field(str)
  filename = Field(str, default=None)
  options = Field(dict, Remainder())

  def canonical(self, parent_directory: str) -> 'DatabaseConfig':
    self = copy.copy(self)
    self.filename = _canonical(self.filename, parent_directory)
    return self

  def bind_kwargs(self) -> dict:
    return {'provider': self.provider, 'filename': self.filename, **self.options}


class DeployConfig(Struct):
  products_install_directory = Field(str, FieldName('products-install-directory'), default='var/products')
  services_install_directory = Field(str, FieldName('services-install-directory'), default='var/services')
  tasks_data_directory = Field(str, FieldName('tasks-data-directory'), default='var/data/tasks')
  pidfile = Field(str, default='var/run/cube-daemon.pid')
  stdout = Field(str, default='var/log/stdout.log')
  stderr = Field(str, default='var/log/stderr.log')

  def canonical(self, parent_directory: str) -> 'DeployConfig':
    self = copy.deepcopy(self)
    self.products_install_directory = _canonical(self.products_install_directory, parent_directory)
    self.services_install_directory = _canonical(self.services_install_directory, parent_directory)
    self.tasks_data_directory = _canonical(self.tasks_data_directory, parent_directory)
    self.pidfile = _canonical(self.pidfile, parent_directory)
    self.stdout = _canonical(self.stdout, parent_directory)
    self.stderr = _canonical(self.stderr, parent_directory)
    return self


class ApplicationConfig(Struct):
  auth = Field(AuthConfig)
  database = Field(DatabaseConfig)
  deploy = Field(DeployConfig, default=Field.DEFAULT_CONSTRUCT)
  server = Field(ServerConfig)

  def canonical(self, parent_directory: str) -> 'ApplicationConfig':
    return ApplicationConfig(
      self.auth,
      self.database.canonical(parent_directory),
      self.deploy.canonical(parent_directory),
      self.server)

  @classmethod
  def load(cls, mapper: ObjectMapper, filename: str, data: str = None) -> 'ApplicationConfig':
    """
    Loads the application config from a file. The config is expected to be
    nested in a "$.runtime" key.
    """

    if data is None:
      with open(filename) as fp:
        data = fp.read()
    data = yaml.safe_load(data)
    return mapper.deserialize(data, {'runtime': Field(cls)}, filename=filename).runtime
